//
// Skirmish Base Script.
//
// contains the rules for starting and ending a game.
// as well as warning messages.
//
// /////////////////////////////////////////////////////////////////

var lastHitTime = 0;
var global=this;

function eventGameInit()
{
	if (typeof(global.scavengerPlayer)=="undefined") // HACK: until scavengerPlayer is exposed to scripts.
		global.scavengerPlayer=Math.max(7,maxPlayers);
	hackNetOff();
	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		// make structures available to build
		enableStructure("A0HardcreteMk1Wall", playnum);
		enableStructure("MiniCC", playnum);
		enableStructure("MiniLightFactory", playnum);
		enableStructure("MiniLaboratory", playnum);
		enableStructure("MiniPowerGenerator", playnum);
		enableStructure("A0ResourceExtractor", playnum);

		// We need to enable these in order for scripts to be able to generate their templates.
		makeComponentAvailable("wheeled01", playnum);
		makeComponentAvailable("V-Tol", playnum);
		makeComponentAvailable("MiniExcavatorArm", playnum);
		makeComponentAvailable("Spade1Mk1", playnum);
		makeComponentAvailable("MiniExcavatorBody", playnum);
		makeComponentAvailable("Body1REC", playnum);
		
		enableResearch("R-MiniMG", playnum);
		enableResearch("R-MiniEngineering", playnum);
		enableResearch("R-MiniAdvancedLaboratory", playnum);
		
		setStructureLimits("MiniCC",1,playnum);
		setStructureLimits("MiniLightFactory",5,playnum);
		setStructureLimits("MiniHeavyFactory",5,playnum);
		setStructureLimits("MiniPowerGenerator",10,playnum);
		setStructureLimits("MiniLaboratory",5,playnum);
		setStructureLimits("MiniAdvancedLaboratory",5,playnum);
		setStructureLimits("MiniRepairFacility",5,playnum);
		setStructureLimits("MiniChopperFactory",5,playnum);
		setStructureLimits("MiniRearmPad",50,playnum);
	}

	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		setPower(1000, playnum);
		var structs = enumStruct(playnum);
		for (var i = 0; i < structs.length; i++)
			removeStruct(structs[i]);
	}
	
	var scavs = enumStruct(scavengerPlayer);
	for (var i = 0; i < scavs.length; i++)
		removeStruct(scavs[i]);

	setMiniMap(false);
	setDesign(false);

	enableTemplate("ConstructionDroid");
	enableTemplate("MiniExcavator");
    enableTemplate("MiniBuggy");
    enableTemplate("MiniCamper");
    enableTemplate("MiniFiretruck");
    enableTemplate("MiniHeavyTruck");
    enableTemplate("MiniJeep");
    enableTemplate("MiniIcevan");
    enableTemplate("MiniMGChopper");
    enableTemplate("MiniPoliceChopper");
    enableTemplate("MiniChinook");
    enableTemplate("MiniSchoolBus");
    enableTemplate("MiniTractor");
    enableTemplate("MiniTrike");
	
	hackNetOn();
	setTimer("checkEndConditions", 100);
}

function getLivingObjectCount(playnum) {
	var s = enumDroid(playnum).length;
	s += enumStruct(playnum, "MiniLightFactory").length;
	s += enumStruct(playnum, "MiniHeavyFactory").length;
	s += enumStruct(playnum, "MiniChopperFactory").length;
	return s;
}

// /////////////////////////////////////////////////////////////////
// END CONDITIONS
function checkEndConditions()
{
	// Losing Conditions
	if (getLivingObjectCount(me) == 0)
	{
		var gameLost = true;

		/* If teams enabled check if all team members have lost  */
		if (alliancesType == ALLIANCES_TEAMS)
		{
			for (var playnum = 0; playnum < maxPlayers; playnum++)
			{
				if (playnum != me && allianceExistsBetween(me, playnum))
				{
					if (getLivingObjectCount(playnum) == 0)
					{
						gameLost = false;	// someone from our team still alive
						break;
					}
				}
			}
		}

		if (gameLost)
		{
			gameOverMessage(false);
			removeTimer("checkEndConditions");
			return;
		}
	}
	
	// Winning Conditions
	var gamewon = true;

	// check if all enemies defeated
	for (var playnum = 0; playnum < maxPlayers; playnum++)
	{
		if (playnum != me && !allianceExistsBetween(me, playnum))	// checking enemy player
		{
			if (getLivingObjectCount(playnum) > 0)
			{
				gamewon = false;	//one of the enemies still alive
				break;
			}
		}
	}

	if (gamewon)
	{
		gameOverMessage(true);
		removeTimer("checkEndConditions");
	}
}

// /////////////////////////////////////////////////////////////////
// WARNING MESSAGES
// Base Under Attack
function eventAttacked(victimObj, attackerObj)
{
	if (gameTime > lastHitTime + 10)
	{
		lastHitTime = gameTime;
		if (victimObj.type == STRUCTURE)
		{
			playSound("pcv337.ogg", victimObj.x, victimObj.y, victimObj.z);	// show position if still alive
		}
		else
		{
			playSound("pcv399.ogg", victimObj.x, victimObj.y, victimObj.z);
		}
	}
}

function eventStructureBuilt(struct)
{
        if (struct.player == selectedPlayer && struct.type == STRUCTURE && struct.stattype == HQ)
        {
                setMiniMap(true); // show minimap
        }
}

function eventDestroyed(victim)
{
        if (victim.player == selectedPlayer && victim.type == STRUCTURE && victim.stattype == HQ)
        {
                setMiniMap(false); // hide minimap if HQ is destroyed
        }
}


